from netmiko import BaseConnection
from napalm.base import models as napalm_models, NetworkDriver

from ..base import SCBaseDriver


class PSIMN(NetworkDriver, SCBaseDriver):
    """
        Very sketchy, basic driver for noc-wan-nokia-psi-dc
        running 1830PSIMN-25.6-0 to pull config backups.

    Netmiko doesn't appear to model this so we're
    doing a paramiko-based driver instead.
    """

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        if optional_args is None:
            optional_args = {}
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.platform = "psimn"

        self.device = None

    def open(self):
        """
        Implementation of NAPALM method 'open' to open a connection to the device.
        """
        self.device = BaseConnection(
            ip=self.hostname,
            username=self.username,
            password=self.password,
            timeout=self.timeout,
        )

        # switching to a different CLI that we know how to do 'show run' on
        self.device.send_command(
            "//\n", expect_string="Switching to MD-CLI", cmd_verify=False
        )

    def close(self):
        """
        Implementation of NAPALM method 'close'. Closes the connection to the device and does
        the necessary cleanup.
        """
        self.device.remote_conn.close()

    def send_command(self, cmd: str) -> str:
        return self.device.send_command(cmd)

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        config = self.send_command("admin show running | no-more")

        return {"startup": {}, "candidate": {}, "running": config}



from typing import List
from napalm.eos import EOSDriver
from napalm.base import models as napalm_models
from netaddr import EUI, mac_unix_expanded
import re

from ..base import SCBaseDriver
from ..models import InventoryDict

from pprint import pprint


class SCEOSDriver(EOSDriver, SCBaseDriver):
    ("fabric_module",)
    ("fan",)
    ("linecard",)
    ("optic",)
    ("psu",)
    ("re",)
    ("stack_cable",)
    ("stack_member",)
    ("uplink_module",)
    ("aoc",)
    ("dac",)

    INVENTORY_TO_TYPE = {
        r"Fabric": "fabric_module",
        r"Linecard": "linecard",
        r"Supervisor": "re",
    }

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Forcing ssh transport since we don't enable the web interface
        """
        optional_args = optional_args if optional_args else {}
        optional_args["transport"] = "ssh"

        super().__init__(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

    def _get_optic_port(self, slot, model):
        """
        Certain ports on certain models need to have a "/1"
        appended to each one.
        """
        if model == "DCS-7060X6-64PE" and int(slot) < 65:
            return f"Ethernet{slot}/1"
        if model == "DCS-7280DR3A-54":
            return f"Ethernet{slot}/1"
        if model == "DCS-7050SX3-48C8C" and int(slot) > 48:
            return f"Ethernet{slot}/1"
        if model == "DCS-7280SR3-40YC6" and int(slot) > 40:
            return f"Ethernet{slot}/1"

        return f"Ethernet{slot}"

    def get_inventory(self) -> List[InventoryDict]:
        inventory = self._run_commands(["show inventory"], encoding="json")

        results = []

        # luckily for us we get the device model with this call
        # which we need to munge the name of the iterface
        # the optic is on
        model = inventory[0]["systemInformation"]["name"]

        ### optics
        for slot, optic in inventory[0]["xcvrSlots"].items():
            if optic.get("modelName"):
                results.append(
                    {
                        "type": "optic",
                        "subtype": optic["modelName"],
                        "name": self._get_optic_port(slot, model),
                        "part_number": optic["modelName"],
                        "serial_number": optic["serialNum"],
                    },
                )

        ### line cards
        for slot, card in inventory[0]["cardSlots"].items():
            if card.get("serialNum"):
                results.append(
                    {
                        "type": self._get_inventory_type(slot),
                        "subtype": card["modelName"],
                        "name": f"Ethernet{slot}",
                        "part_number": card["modelName"],
                        "serial_number": card["serialNum"],
                    },
                )

        ### PSUs
        for slot, psu in inventory[0]["powerSupplySlots"].items():
            if psu.get("serialNum"):
                results.append(
                    {
                        "type": "psu",
                        "subtype": None,
                        "name": f"PSU {slot}",
                        "part_number": psu["name"],
                        "serial_number": psu["serialNum"],
                    },
                )

        ### FANs
        for slot, fan in inventory[0]["fanTraySlots"].items():
            if fan.get("serialNum"):
                results.append(
                    {
                        "type": "fan",
                        "subtype": None,
                        "name": f"FAN {slot}",
                        "part_number": fan["name"],
                        "serial_number": fan["serialNum"],
                    },
                )
        return results

    def get_ipv6_neighbors_table(self) -> List[napalm_models.IPV6NeighborDict]:
        """
        Not implemented in Napalm base EOS!
        """
        output = []
        neighbors = self._run_commands(["show ipv6 neighbors vrf all"], encoding="json")
        if not neighbors:
            return output

        for _, vrf in neighbors[0]["vrfs"].items():
            for entry in vrf["ipV6Neighbors"]:
                mac = EUI(entry["hwAddress"])
                mac.dialect = mac_unix_expanded
                output.append(
                    {
                        "interface": entry["interface"],
                        "mac": str(mac),
                        "ip": entry["address"],
                        "age": entry["age"],
                        "state": "REACH",
                    }
                )
        return output

    def get_optics(self):
        """
        napalm eos driver is problematic in several ways so we get to rewrite it
        """
        command = ["show interfaces transceiver"]

        output = self._run_commands(command, encoding="json")[0]["interfaces"]

        # Formatting data into return data structure
        optics_detail = {}

        for port, port_values in output.items():
            if not port_values:
                continue

            # multi-channel optics show up as separate ports, eg
            # Ethernet45/1, Ethernet45/2, Ethernet45/3. We're going to
            # use the 'slot' attribute instead
            port_name = port_values["slot"] + "/1   "
            if port_name not in optics_detail:
                optics_detail[port_name] = {"physical_channels": {"channel": []}}

            # Defaulting avg, min, max values to 0.0 since device does not
            # return these values
            optic_states = {
                "index": port_values["channel"],
                "state": {
                    "input_power": {
                        "instant": (
                            port_values["rxPower"] if "rxPower" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                    "output_power": {
                        "instant": (
                            port_values["txPower"] if "txPower" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                    "laser_bias_current": {
                        "instant": (
                            port_values["txBias"] if "txBias" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                },
            }

            optics_detail[port_name]["physical_channels"]["channel"].append(
                optic_states
            )

        return optics_detail



import os
import pytest
import json
from pathlib import Path

from mock_get import get_mock_driver

test_dir = f"{os.getcwd()}/test/mock_data"
test_paths = []
for platform in os.listdir(test_dir):
    for getter in os.listdir(f"{test_dir}/{platform}"):
        for test in os.listdir(f"{test_dir}/{platform}/{getter}"):
            test_paths.append(Path(f"{test_dir}/{platform}/{getter}/{test}"))


@pytest.mark.parametrize("test_path", test_paths)
def test_getter(test_path: Path):
    driver = get_mock_driver(test_path.parts[-3])
    with driver(test_path) as device:
        result = getattr(device, test_path.parts[-2])()

    with open(f"{test_path}/expected_result.json", encoding="utf-8") as fh:
        expected_result = json.load(fh)

    assert result == expected_result



from typing import Optional, Dict
from napalm.base import NetworkDriver


class NokiaSRLSSHDriver(NetworkDriver):
    """
    Netmiko based Nokia SRL driver :(
    """

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        timeout: int = 60,
        optional_args: Optional[Dict] = None,
    ) -> None:
        if optional_args is None:
            optional_args = {}
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout

    def open(self):
        self.device = self._netmiko_open("nokia_srl")

    def close(self):
        self._netmiko_close()

    def send_command(self, cmd: str) -> str:
        """
        Sends command with netmiko and returns the result
        """
        return self.device.send_command(cmd)

    def get_config(self, retrieve="all", full=False, sanitized=False, format="text"):
        result = {"startup": "", "candidate": "", "running": ""}
        if retrieve in ["all", "running"]:
            result["running"] = self.send_command("info from running flat")
        if retrieve in ["all", "startup"]:
            result["startup"] = self.send_command("info from startup flat")
        if retrieve in ["all", "candidate"]:
            result["candidate"] = self.send_command("info from candidate flat")
        return result



from typing import Dict, List
import re
from napalm.base import models as napalm_models
from netmiko import ConnectHandler

from . import models as sc_models


class SCBaseDriver:
    """
    Base class that all children should inherit
    """

    # populate in child classes
    INVENTORY_TO_TYPE = {}

    def _get_inventory_type(self, name: str) -> str:
        """
        Maps the name of the inventory item to its type
        """
        for pattern, inv_type in self.INVENTORY_TO_TYPE.items():
            if inv_type and inv_type not in sc_models.VALID_INVENTORY_TYPES.__args__:
                raise TypeError(f"Invalid Inventory type {inv_type}")
            if re.search(pattern, name):
                return inv_type

        raise ValueError(f"Unknown inventory item {name}")

    def get_ipv6_neighbors_table(self) -> List[napalm_models.IPV6NeighborDict]:
        raise NotImplementedError

    def get_arp_table(self) -> List[napalm_models.ARPTableDict]:
        """napalm get arp table"""
        raise NotImplementedError

    def get_config(self) -> napalm_models.ConfigDict:
        """napalm get config"""
        raise NotImplementedError

    def get_facts(self) -> napalm_models.FactsDict:
        """napalm get facts"""
        raise NotImplementedError

    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        """napalm get lldp neighbors"""
        raise NotImplementedError

    def get_interfaces(self) -> Dict[str, napalm_models.InterfaceDict]:
        """ " napalm get interfaces"""
        raise NotImplementedError

    def get_inventory(self) -> List[sc_models.InventoryDict]:
        """sc get inventory"""
        raise NotImplementedError

    def get_mac_address_table(self) -> List[napalm_models.MACAdressTable]:
        """napalm get mac address table"""
        raise NotImplementedError

    def get_optics(self) -> Dict[str, napalm_models.OpticsDict]:
        """napalm get optics"""
        raise NotImplementedError


class SCBaseNetconfDriver(SCBaseDriver):
    """
    Inclues some helper xml functions
    """

    netmiko_host_type = "linux"

    def ssh_conn(self) -> ConnectHandler:
        """
        Ugly workaround for getting stuff via the cli over ssh.
        Starts a netmiko ssh connection handler and returns it,
        allowing you to interact with the CLI of the device.
        """
        optional_args = {}
        if getattr(self, "optional_args", False):
            optional_args = self.optional_args if self.optional_args else {}

        args = {
            "device_type": self.netmiko_host_type,
            "host": self.hostname,
            "username": self.username,
            "password": self.password,
            "ssh_config_file": optional_args.get("ssh_config", None),
        }
        return ConnectHandler(**args)

    def _find_txt(self, xml_tree, path, default=None, namespaces=None):
        """
        Stolen from the napalm iosxr driver
        Extract the text value from a leaf in an XML tree using XPath.

        Will return a default value if leaf path not matched.
        :param xml_tree:the XML Tree object. <type'lxml.etree._Element'>.
        :param path: XPath to be applied in order to extract the desired data.
        :param default:  Value to be returned in case of a no match.
        :param namespaces: namespace dictionary.
        :return: a str value or None if leaf path not matched.
        """
        value = None
        xpath_applied = xml_tree.xpath(path, namespaces=namespaces)

        if xpath_applied:
            if not len(xpath_applied[0]):
                if xpath_applied[0].text is not None:
                    value = xpath_applied[0].text.strip()
                else:
                    value = ""
        else:
            value = default

        return value

    # Helper xml methods that always pass in our namespaces by default
    def _text(self, xml_tree, path, default=None):
        return self._find_txt(xml_tree, path, default, namespaces=self.NS)

    def _xpath(self, xml_tree, path):
        return getattr(xml_tree, "xpath")(path, namespaces=self.NS)

    def _find(self, xml_tree, element):
        return getattr(xml_tree, "find")(element, namespaces=self.NS)

    def _iterfind(self, xml_tree, element):
        return getattr(xml_tree, "iterfind")(element, namespaces=self.NS)



import argparse
from pathlib import Path
import json

import custom_napalm
from mock_drivers import *

from pprint import pprint

PLATFORM_MAP = {
    "iosxr": MockIOSXR,
    "nxos": MockNXOS,
    "junos": MockJunos,
    #    "sros": SCNokiaSROSDriver,
    "srl": MockNokiaSRL,
    "eos": MockEOS,
}


def get_mock_driver(platform: str):
    """
    Returns network driver based on platform string.
    """
    for valid_platform, driver in PLATFORM_MAP.items():
        if valid_platform == platform:
            return driver

    raise NotImplementedError(f"Unsupported platform {platform}")


if __name__ == "__main__":
    MOCK_DATA_DIR = Path(__file__).parent / "mock_data"

    parser = argparse.ArgumentParser(
        description="""Generate mock output data based on raw router input saved in the mock_data directory"""
    )
    parser.add_argument("platform", choices=custom_napalm.PLATFORM_MAP)
    parser.add_argument("getter", type=str)
    parser.add_argument("test_name", type=str)
    parser.add_argument(
        "--save", action="store_true", help="Save result to test folder"
    )

    args = parser.parse_args()

    test_dir = MOCK_DATA_DIR / args.platform / args.getter / args.test_name

    folder = Path(test_dir)
    if not folder.exists():
        raise FileNotFoundError(f"No test folder {test_dir}")

    driver = get_mock_driver(args.platform)
    with driver(test_dir) as device:
        result = getattr(device, args.getter)()

    pprint(result)

    if args.save:
        with open(f"{test_dir}/expected_result.json", "w", encoding="utf-8") as fh:
            fh.write(json.dumps(result, indent=4))



from napalm_sros import NokiaSROSDriver
from napalm.base import models as napalm_models

from ncclient.xml_ import to_ele

from ..base import SCBaseNetconfDriver, sc_models
from .nc_filters import GET_INVENTORY


class SCNokiaSROSDriver(NokiaSROSDriver, SCBaseNetconfDriver):
    netmiko_host_type = "alcatel_sros"

    NS = {
        "state": "urn:nokia.com:sros:ns:yang:sr:state",
    }

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        if format != "text":
            return super().get_config(
                retrieve=retrieve, full=full, sanitized=sanitized, format=format
            )

        config = {"startup": "", "running": "", "candidate": ""}

        with self.ssh_conn() as ssh_device:
            config["running"] = ssh_device.send_command("admin show configuration flat")
            config["running"] += ssh_device.send_command(
                "admin show configuration bof flat"
            )

        return config

    def get_inventory(self) -> list[sc_models.InventoryDict]:
        inv_xml = to_ele(
            self.conn.get(
                filter=GET_INVENTORY["_"], with_defaults="report-all"
            ).data_xml
        )
        # print(etree.tostring(result, pretty_print=True, encoding="unicode"))

        results = []
        for i in self._xpath(inv_xml, ".//state:port"):
            if not self._xpath(i, ".//state:transceiver"):
                continue

            results.append(
                {
                    "type": "optic",
                    "subtype": self._text(i, ".//state:vendor-part-number").strip(),
                    "name": self._text(i, ".//state:port-id"),
                    "part_number": self._text(i, ".//state:model-number").strip(),
                    "serial_number": self._text(i, ".//state:vendor-serial-number"),
                }
            )

        return results

    def get_optics(self):
        """
        Filtering out ports that don't have optics
        """
        results = super().get_optics()
        better_results = {}
        for i_name, data in results.items():
            if data["physical_channels"]["channel"]:
                better_results[i_name] = data
        return better_results
